package PL::FirstFollow;
use strict;
use warnings;

use Set::Scalar;
use Grammar;
use Carp;

require Exporter;

our @ISA = qw(Exporter);
our @EXPORT_OK = qw( slurp compute_grammar first follow );
our $VERSION = '0.02';

####################################################################
# Usage      : ????
# Purpose    : ????
# Returns    : ????
# Parameters : ????
# Throws     : no exceptions
# Comments   : none
# See Also   : n/a

sub slurp  {
  my $filename = shift;

  local $/ = undef;
  open(my $FILE, $filename) or croak "can't open $filename. Current directory contains:\n<<".glob('./*').">>";
  my $g;
  $$g = <$FILE>;
  close $FILE;
  return $g;
}

####################################################################
# Usage      : ????
# Purpose    : ????
# Returns    : ????
# Parameters : ????
# Throws     : no exceptions
# Comments   : none
# See Also   : n/a

sub compute_grammar {
  my $grammar = shift;
  my $epsilon = shift || '<EPSILON>';
  my $eoi = shift || '$';
  
  $grammar->{EPSILON} = $epsilon;
  $grammar->{EOI} = $eoi;
  initialize($grammar);
  $grammar->{NULLABLE} = is_nullable($grammar);
  $grammar->{FIRST} = _first($grammar);
  $grammar->{FOLLOW} = follow($grammar);
  
  return $grammar;
}


####################################################################
# Usage      : ????
# Purpose    : ????
# Returns    : ????
# Parameters : ????
# Throws     : no exceptions
# Comments   : none
# See Also   : n/a

sub initialize {
  my $grammar = shift;

  my %first;
  my %follow;

  $grammar->{termset} = Set::Scalar->new(@{$grammar->{TERM}});
  $grammar->{ntermset} = Set::Scalar->new(keys %{$grammar->{NTERM}});
  my @nterm = keys %{$grammar->{NTERM}};   # vector de no terminales
  foreach my $var (@nterm) {
    $first{$var} = Set::Scalar->new;
    $follow{$var} = Set::Scalar->new;
  }
  $grammar->{FIRST} = \%first;
  $grammar->{FOLLOW} = \%follow;
}

####################################################################
# Usage      : ????
# Purpose    : ????
# Returns    : ????
# Parameters : ????
# Throws     : no exceptions
# Comments   : none
# See Also   : n/a

sub _first {
  my $grammar = shift;

  my $null = $grammar->{NULLABLE};
  my %first = %{$grammar->{FIRST}};

  foreach my $rule (@{$grammar->{RULES}}) {
    # parte izquierda
    my $var =  @{$rule}[0];

    # parte derecha
    foreach my $s (@{@{$rule}[1]}) {
      # simbolo es terminal o variable no nula
      $first{$var}->insert($s);

      last unless $null->has($s);
    } #foreach s
    $first{$var}->delete($var);                    # evitar var en first{var}
  } #foreach rule

  _replace($grammar->{NTERM}, \%first);

  # insertar epsilon en los anulables
  foreach my $v (keys(%{$grammar->{NTERM}})) {
      if ($null->has($v)) { $first{$v}->insert($grammar->{EPSILON}) };
  }
  return \%first;
}

####################################################################
# Usage      : ????
# Purpose    : ????
# Returns    : ????
# Parameters : ????
# Throws     : no exceptions
# Comments   : none
# See Also   : n/a

sub follow {
  my $grammar = shift;

  my %first = %{$grammar->{FIRST}};
  my %follow = %{$grammar->{FOLLOW}};
  my $nterm = $grammar->{ntermset};

  $follow{$grammar->{START}}->insert($grammar->{EOI});
  
  foreach my $rule (@{$grammar->{RULES}}) {
    # parte izquierda
    my $var =  @{$rule}[0];

    # parte derecha
    my @alt = @{@{$rule}[1]};
    my $i = 0;
    for (my $i = 0; $i < @alt; $i++) {
      my $s = $alt[$i];
      if ($nterm->has($s)) {
        if ($i != $#alt) {                             # es el ultimo simbolo?
          my @tmp_alt = @alt;
          my @beta = splice @tmp_alt, $i+1, @tmp_alt-$i-1;  # beta=[s+1..fin]
          my $first_beta = first($grammar, @beta);
          $follow{$s} = $follow{$s} + $first_beta;
          
          if ($first_beta->has($grammar->{EPSILON})) {              # beta es anulable? 
            $follow{$s}->insert($var);
          }

          $follow{$s}->delete($grammar->{EPSILON});                 # evitar epsilon en follow{s}
        } 
        else { 
          $follow{$s}->insert($var);                    # si es ultimo simbolo incluir follow{var}  
        }
        $follow{$s}->delete($s);                        # evitar s en follow{s}
      }
    } #for s
  } #foreach rule
  
  _replace($grammar->{NTERM}, \%follow);
  return \%follow;
}

####################################################################
# Usage      : is_nullable($grammar)
# Returns    : El Scalar::Set { A in V / A =*=> epsilon }
# Parameters : La E.D. de la gramatica generada por el modulo Grammar
# Throws     : no exceptions

sub is_nullable {
  my $grammar = shift;

  my $null = Set::Scalar->new(keys %{$grammar->{NULL}});
  my $change;
  do {
    $change = 0;
    foreach my $rule (@{$grammar->{RULES}}) {
      my $var =  @{$rule}[0];
    
      my $is_null = 0;
      if (not $null->has($var)) {                 # si ya es anulable saltarla
        foreach my $s (@{@{$rule}[1]}) {
          ?? ???????????????? ? ???????? ? ?? ?
          ???? ? ???????? ? ?? ????? ?           
        } #foreach s
      }
      if ($is_null) { $null->insert($var); $change = 1; }
    } #foreach rule
  } while ($change);
  return $null;
}

####################################################################
# Usage      : ????
# Purpose    : ????
# Returns    : ????
# Parameters : ????
# Throws     : no exceptions
# Comments   : Warning! Es un requisito que el conjunto $f->{$var} no contiene a $var
# See Also   : n/a

sub _replace {
  my ($nterm, $f) = @_;
  my $change;
  my %nterm = %$nterm;   # vector de no terminales
  my @nterm = keys(%nterm);
  do {
    $change = 0;
    foreach my $var (@nterm) {
      # para cada e del cto $f{var}
      my @members = $f->{$var}->members;
      foreach my $e (@members) {
        if (exists($nterm{$e})) {
          $f->{$var} = $f->{$var} + $f->{$e};
          $f->{$var}->delete($e);
          $f->{$var}->delete($var);
          $change = 1;
        }
      }
    }
  } while ($change);
}

####################################################################
# Usage      : ????
# Purpose    : ????
# Returns    : ????
# Parameters : ????
# Throws     : no exceptions
# Comments   : none
# See Also   : n/a

sub first {
  my $grammar = shift;
  my @str = @_;

  my $null = $grammar->{NULLABLE};
  my $first = $grammar->{FIRST};
  my $term = $grammar->{termset};
  my %NTERM = %{$grammar->{NTERM}};
  my $result = Set::Scalar->new;
  my $is_null = 1;
  ??????? ?? ?? ?????? ?
    ?? ???????????????? ? 
      ????????????????????
      ???????? ? ??
      ?????
    ?
    ????? ???????????????????? ?
      ??????? ? ??????? ? ?????????????
      ?? ???? ??????????????? ?
        ???????? ? ??
        ?????
      ?
    ?
  ? 
  if ($is_null) { $result->insert($grammar->{EPSILON}); }
  else { $result->delete($grammar->{EPSILON}); }

  return $result;
}

1;

__END__

=head1 NAME
 
<Module::Name> - <One line description of module's purpose>
 
 
=head1 VERSION
 
The initial template usually just has:
 
This documentation refers to <Module::Name> version 0.0.1
 
 
=head1 SYNOPSIS
 
    use <Module::Name>;
    # Brief but working code example(s) here showing the most common usage(s)
 
    # This section will be as far as many users bother reading
    # so make it as educational and exemplary as possible.
  
  
=head1 DESCRIPTION
 
A full description of the module and its features.
May include numerous subsections (i.e. =head2, =head3, etc.) 
 
 
=head1 SUBROUTINES/METHODS 
 
A separate section listing the public components of the module's interface. 
These normally consist of either subroutines that may be exported, or methods
that may be called on objects belonging to the classes that the module provides.
Name the section accordingly.
 
In an object-oriented module, this section should begin with a sentence of the 
form "An object of this class represents...", to give the reader a high-level
context to help them understand the methods that are subsequently described.
 
 
=head1 DIAGNOSTICS
 
A list of every error and warning message that the module can generate
(even the ones that will "never happen"), with a full explanation of each 
problem, one or more likely causes, and any suggested remedies.
 
 
=head1 CONFIGURATION AND ENVIRONMENT
 
A full explanation of any configuration system(s) used by the module,
including the names and locations of any configuration files, and the
meaning of any environment variables or properties that can be set. These
descriptions must also include details of any configuration language used.
 
 
=head1 DEPENDENCIES
 
A list of all the other modules that this module relies upon, including any
restrictions on versions, and an indication whether these required modules are
part of the standard Perl distribution, part of the module's distribution,
or must be installed separately.
 
 
=head1 INCOMPATIBILITIES
 
A list of any modules that this module cannot be used in conjunction with.
This may be due to name conflicts in the interface, or competition for 
system or program resources, or due to internal limitations of Perl 
(for example, many modules that use source code filters are mutually 
incompatible).
 
 
=head1 BUGS AND LIMITATIONS
 
A list of known problems with the module, together with some indication
whether they are likely to be fixed in an upcoming release.
 
Also a list of restrictions on the features the module does provide: 
data types that cannot be handled, performance issues and the circumstances
in which they may arise, practical limitations on the size of data sets, 
special cases that are not (yet) handled, etc.
 
The initial template usually just has:
 
There are no known bugs in this module. 
Please report problems to <Maintainer name(s)>  (<contact address>)
Patches are welcome.
 
=head1 AUTHOR
 
<Author name(s)>  (<contact address>)
 
 
=head1 LICENCE AND COPYRIGHT
 
Copyright (c) <year> <copyright holder> (<contact address>). All rights reserved.
 
Followed by whatever licence you wish to release it under. 
For Perl code that is often just:
 
This module is free software; you can redistribute it and/or
modify it under the same terms as Perl itself. See L<perlartistic>.
 
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 

