### Test para FirstFollow
### make test TEST_VERBOSE=1
use strict;
use Test::More tests => 83;

BEGIN { use_ok('PL::FirstFollow', qw(slurp compute_grammar first follow) ) };

my (%a, %b, %x, %y, %i, %o);
my $g;

############################################################################### 
sub make_test {
  my $filename = shift;
  my $g = slurp($filename);
  my $grammar = Grammar::Parse($$g);
  compute_grammar($grammar);

  %a    = %{$grammar->{FIRST}};
  %b    = %{$grammar->{FOLLOW}};
  foreach my $v (keys %a) {
    ok($a{$v} == $i{$v}, "$filename first{$v}");
    ok($b{$v} == $o{$v}, "$filename follow{$v}");
  }
}

sub set {
  return Set::Scalar->new(@_);
}



############################################################################### 
#                          TEST PARA FIRST Y FOLLOW
############################################################################### 

$i{A}       = set(qw('a'));
$i{B}       = set(qw('c' <EPSILON>));
$o{A}       = set(qw('b' 'c' $));
$o{B}       = set(qw('b' 'c' $));
make_test('t/yacc/g1.yp');



############################################################################### 

$i{S}       = set(qw('a' <EPSILON>));
$o{S}       = set(qw($ 'b'));
make_test('t/yacc/aSb.yp');



############################################################################### 

$i{input}   = set(qw(NUM VAR '\n' error '-' '\(' <EPSILON>) );
$i{exp}     = set(qw(NUM VAR '-' '\('));
$i{line}    = set(qw(NUM VAR '\n' error '-' '\('));
$o{input}   = set(qw(NUM VAR '\n' error '-' '\(' $));
$o{exp}     = set(qw('\n' '+' '-' '*' '/' '^' '\)'));
$o{line}    = set(qw(NUM VAR '\n' error '-' '\(' $));
make_test('t/yacc/Calc.yp');



############################################################################### 

$i{list}    = set(qw('\n' NUMBER <EPSILON>));
$i{e}       = set(qw(NUMBER));
$o{list}    = set(qw($ '\n' NUMBER));
$o{e}       = set(qw('\n' NUMBER '&' '@' $));
make_test('t/yacc/Precedencia.yp');



############################################################################### 

$i{S}       = set(qw('a' 'b' 'c'));
$i{A}       = set(qw('a' <EPSILON>));
$i{B}       = set(qw('b' <EPSILON>));
$i{C}       = set(qw('a' 'b'));
$i{D}       = set(qw('a' 'b' <EPSILON>));
$o{S}       = set(qw($));
$o{A}       = set(qw('a' 'b' 'c' $));
$o{B}       = set(qw('a' 'b' 'c' $));
$o{C}       = set(qw('a' 'b' $));
$o{D}       = set(qw($));
make_test('t/yacc/ex1.yp');



############################################################################### 

$i{S}       = set(qw('b' 'c' 'd' 'a' 'g'));
$i{B}       = set(qw('b' <EPSILON>));
$i{C}       = set(qw('c' 'd' 'a'));
$i{E}       = set(qw('g' 'c'));
$i{D}       = set(qw('d' <EPSILON>));
$o{S}       = set(qw($ 'f'));
$o{B}       = set(qw('c' 'd' 'a' $ 'f' 'g'));
$o{C}       = set(qw('d' 'g' 'c'));
$o{E}       = set(qw('c' 'd' 'a' $ 'f' 'g'));
$o{D}       = set(qw('b' $ 'f' 'g' 'c' 'a'));
make_test('t/yacc/ex2.yp');



############################################################################### 

$i{program}     = set(qw(';' 'INT' 'STRING' 'ID' 'P' <EPSILON>));
$i{declarations}= set(qw('INT' 'STRING'));
$i{declaration} = set(qw('INT' 'STRING'));
$i{statements}  = set(qw(';' 'ID' 'P' <EPSILON>));
$i{statement}   = set(qw('ID' 'P' <EPSILON>));
$i{expression}  = set(qw('ID' '\(' 'NUM' 'STR'));
$i{term}        = set(qw('ID' '\(' 'NUM' 'STR'));
$i{factor}      = set(qw('ID' '\(' 'NUM' 'STR'));
$i{idlist}      = set(qw('ID'));
$o{program}     = set(qw($));
$o{declarations}= set(qw(';' 'ID' 'P' $));
$o{declaration} = set(qw(';'));
$o{statements}  = set(qw($));
$o{statement}   = set(qw(';' $));
$o{expression}  = set(qw(';' '\)' $));
$o{term}        = set(qw(';' '+' '\)' $));
$o{factor}      = set(qw(';' '+' '*' '\)' $));
$o{idlist}      = set(qw(';'));
make_test('t/yacc/tutu.yp');



############################################################################### 

$i{program}       = set(qw(';' 'INT' 'STRING' 'ID' 'P' <EPSILON>));
$i{declarations}  = set(qw('INT' 'STRING'));
$i{declarations2} = set(qw('INT' 'STRING' <EPSILON>));
$i{declaration}   = set(qw('INT' 'STRING'));
$i{statements}    = set(qw(';' 'ID' 'P' <EPSILON>));
$i{statements2}   = set(qw(';' <EPSILON>));
$i{statement}     = set(qw('ID' 'P' <EPSILON>));
$i{expression}    = set(qw('ID' '\(' 'NUM' 'STR'));
$i{expression2}   = set(qw('+' <EPSILON>));
$i{term}          = set(qw('ID' '\(' 'NUM' 'STR'));
$i{term2}         = set(qw('*' <EPSILON>));
$i{factor}        = set(qw('ID' '\(' 'NUM' 'STR'));
$i{idlist}        = set(qw('ID'));
$i{idlist2}       = set("<EPSILON>", "','");
$o{program}       = set(qw($));
$o{declarations}  = set(qw(';' 'ID' 'P' $));
$o{declarations2} = set(qw(';' 'ID' 'P' $));
$o{declaration}   = set(qw(';'));
$o{statements}    = set(qw($));
$o{statements2}   = set(qw($));
$o{statement}     = set(qw(';' $));
$o{expression}    = set(qw(';' '\)' $));
$o{expression2}   = set(qw(';' '\)' $));
$o{term}          = set(qw(';' '+' '\)' $));
$o{term2}         = set(qw(';' '+' '\)' $));
$o{factor}        = set(qw(';' '+' '*' '\)' $));
$o{idlist}        = set(qw(';'));
$o{idlist2}       = set(qw(';'));
make_test('t/yacc/tutu-modif.yp');
